import numpy as np
import algorithms.shared as shared
from itertools import combinations
import multiprocessing as mp
from time import time 
from algorithms.util import find_obj,wrapper_parallel

def create_W(irrelevance=None,A=None):
    """Function to create the similarity matrix for the optimization problem.

    Args:
        irrelevance (list, optional): irrelevance scores for each item. Defaults to None.
        A (matrix, optional): Data matrix, if None, then taken from shared memory. Defaults to None.

    Returns:
        W: square matrix of similarities (plus irrelevances if present)
    """
    if A is None:
        A = shared.A
    N = A.shape[0]
    W = A@A.T
    if irrelevance is not None:    
        # W should be rho(u)+rho(v)+2*lam*d_uv     
        W *= 2
        ## add rho(i) to each column and row
        for i,irr in enumerate(irrelevance):
            W[i]+=irr
            W[:,i]+=irr
    return W

def edge_greedy_naive(k,W=None):
    """Edge greedy algorithm proposed by Gollapudi et al.

    Args:
        k (int): cardinality constraint
        W (matrix, optional): Similarity + irrelevance matrix, if None, taken from shared memory. Defaults to None.

    Returns:
        solution: Indicator vector of 1s and 0s
    """
    if W is None:
        W = shared.W
    N = W.shape[0]
    # add inf to not consider diagonals
    W = W + np.diag(np.inf*np.ones(W.shape[0]))
    S = set()
    # D = sim2dist_vec(W)
    # D = D - np.diag(np.diag(D))
    for i in range(k // 2):
        u, v = np.unravel_index(np.argmin(W), W.shape)
        S.add(u)
        S.add(v)
        W[u, :] = np.inf
        W[:, u] = np.inf
        W[v, :] = np.inf
        W[:, v] = np.inf

    if k % 2 != 0:
        other_nodes = list(set(range(N)) - S)
        rand_node = np.random.choice(other_nodes)
        S.add(rand_node)
    
    indicator = np.zeros(N)
    indicator[list(S)] = 1
    return indicator

#! Very slow for large data matrices
def edge_greedy_large(k):
    """Attempt at scaling up edge greedy for large data matrices.
    Still very slow for large matrices.

    Args:
        k (int): cardinality constraint

    Returns:
        solution: Indicator vector of 1s and 0s
    """
    A = shared.A
    # print("Grredy A local: ",id(A))
    N = A.shape[0]
    S = set()
    U = set(range(N))
    for i in range(k // 2):
        min_val = np.inf
        min_ind = None
        for u, v in combinations(U - S, 2):
            sim = A[u] @ A[v]
            if sim < min_val:
                min_val = sim
                min_ind = (u, v)
        S.update(min_ind)

    if k % 2 != 0:
        rand_node = np.random.choice(list(U - S))
        S.add(rand_node)

    indicator = np.zeros(N)
    indicator[list(S)] = 1
    return indicator

def get_sim(edge):
    A = shared.A
    u,v=edge
    return (u,v,A[u] @ A[v])

#! Much slower due to parallel overhead
def edge_greedy_large_parallel(k):
    """Attempt at parallelizing the large version of the edge greedy algorithm
    The overhead in parallel computation seems to be too much

    Args:
        k (int): cardinality

    Returns:
        solution: Indicator vector of 1s and 0s
    """
    A = shared.A
    # print("Greedy A large local: ",id(A))
    N = A.shape[0]
    S = set()
    U = set(range(N))
    for i in range(k // 2):
        v = []
        with mp.Pool(processes=mp.cpu_count()) as pool:
            for o in pool.map(get_sim, combinations(U - S, 2)):
                v.append(o)
            min_v = min(v, key=lambda x: x[2])
        S.update(min_v[:2])

    if k % 2 != 0:
        rand_node = np.random.choice(list(U - S))
        S.add(rand_node)

    indicator = np.zeros(N)
    indicator[list(S)] = 1
    return indicator

def get_edge_greedy_stat(k,A=None,share=True):
    # check that the shared matrix is created
    start = time()
    flag = None
    if shared.W is None and share :
        W = create_W(A=A)
        shared.update_W(W)
    elif A is not None and not share:
        W = create_W(A=A)
        flag = W
    w_creation_time = time()-start

    start = time()
    a = edge_greedy_naive(k, W=flag)
    run_time = time() - start
    assert np.allclose(sum(a), k)

    start = time()
    obj = find_obj(a,A=A)
    obj_time = time()-start

    total_time = w_creation_time+run_time + obj_time
    stat = {
        "algo": "Greedy_A",
        "k": k,
        "min_obj": obj,
        "w_creation_time":w_creation_time,
        "run_time": run_time,
        "obj_time": obj_time,
        "total_time": total_time,
        "best_sol":a,
    }
    return stat


def get_edge_greedy_irr_stat(k,irrelevance,lam=1,A=None,share=True):
    # check that the shared matrix is created
    flag = None

    # scale irrelevances by lambda
    irrelevance = irrelevance*lam
    if shared.W is None and share :
        W = create_W(irrelevance,A)
        shared.update_W(W)
    elif A is not None and not share:
        W = create_W(irrelevance,A)
        flag = W
    
    start = time()
    a = edge_greedy_naive(k, W=flag)
    run_time = time() - start
    assert np.allclose(sum(a), k)

    start = time()
    obj = find_obj(a,irrelevance,A=A)
    obj_time = time()-start

    total_time = run_time + obj_time
    stat = {
        "algo": "Greedy_A_irr",
        "k": k,
        "min_obj": obj,
        "run_time": run_time,
        "obj_time": obj_time,
        "loop_time":0,
        "total_time": total_time,
        "best_sol":a,
        "irrelevance":irrelevance,
        "lam":lam
    }
    return stat