# %%
from mosek.fusion import *
import numpy as np
from time import time

#load shared variables
import algorithms.shared as shared
from algorithms.util import find_obj
from algorithms.rounding import get_eps_delta_solutions

def solve_QP(k, starting_sol=[], irrelevance=None,sparse=False,A=None):
    """Solves  the relaxed QP as stated in Problem 2. of the paper

    
    Args:
        A (numpy array): Data matrix with unit normalized rows of dimensions (N,d)
        k (int) : the cardinality of the solution
        starting_sol (numpy array) : indices of items that must be included in solution
        irrelevance (numpy array) : list of irrelevance scores for each item, dimension (N,)

        
    Returns:
        numpy array: The solution to the QP of dimension (N,)
    """
    if A is None :
        A = shared.A
    # print("solve QP local A: ",id(A)," shape: ",A.shape)
    if sparse:
        A = Matrix.sparse(A)
    # decompose W = (AA^T - I) to get W_d = L^TL
    # use L instead of A^T in objective
    if isinstance(A, Matrix):
        N = A.numRows()
    else:
        N = A.shape[0]
    with Model("QP solver") as M:

        # M = Model("example model")

        x = M.variable(N)

        t = M.variable()

        # constraint 1^T x >= k
        M.constraint("cardinality", Expr.dot(np.ones(N), x), Domain.greaterThan(k))

        # constraint Ix >=0 => -Ix <= 0
        M.constraint("non-negative", x, Domain.greaterThan(0))

        # constraining Ix >= -1 => Ix <= 1
        M.constraint("probability", x, Domain.lessThan(1))

        if len(starting_sol):
            for i in starting_sol:
                M.constraint(f"fixed_{i}", x.index(i), Domain.equalsTo(1))

        # t >= \|A^Tx\|_2^2
        objective = Expr.vstack(0.5, t, Expr.mul(A.transpose(), x))

        # (0.5,t,A^Tx) \in Q_r^(N+2)
        M.constraint(objective, Domain.inRotatedQCone())

        if irrelevance is None:
            # keep t as objective
            M.objective(ObjectiveSense.Minimize, t)
        else:
            # Add irrelevance as linear term t+x^Tc
            M.objective(
                ObjectiveSense.Minimize,
                Expr.add(t, Expr.mul(x.transpose(), irrelevance)),
            )

        M.solve()

        solution = x.level()

    return solution


def get_minsumsim_stat(k, starting_sol=[], sparse=False, eps=1e-2, delta=1e-2, irrelevance=None,A=None,lam=1):    
    """gets the statistics for the minsimsum algorithm
    
    Args:
        k (int): cardinality of the required solution
        starting_sol (list, optional): elements indices that are fixed in solution. Defaults to [].
        sparse (bool, optional): if the data matrix is sparse. Defaults to False.
        eps (float): the parameter to control quality of solution. Defaults to 1e-2
        delta (float): parameter to control quality of solution. Defaults to 1e-2
        irrelevance (list): irrelevance scores for items. Defaults to None
        A (matrix): The data matrix to run the algorithm on. If None matrix stored in shared memory is used. Defaults to None
        lam (float): the tradeoff parameter between irrelevance and sum of similarities. Defaults to 1
    Returns:
        stat : dictionary containing the objective and running times of the algorithm
    """
    if irrelevance is not None:
        # scale irrelevances according to lambda
        irrelevance = irrelevance*lam
        # irrelevance -=1

    start = time()
    z = solve_QP(k, starting_sol=starting_sol,irrelevance=irrelevance,sparse=sparse,A=A)
    qp_time = time() - start

    assert np.allclose(sum(z), k)
    
    start = time()
    possible_sols = get_eps_delta_solutions(z, k, parallel=True,eps=eps,delta=delta)
    rounding_time = time() - start

    start = time()
    obj = [(sol, find_obj(sol,irrelevance=irrelevance,A=A)) for sol in possible_sols]
    best_sol = min(obj, key=lambda x: x[1])

    sol_search_time = time() - start
    total_time = qp_time + rounding_time + sol_search_time
    stat = {
        "algo": "MinSimSum",
        "k": k,
        "min_obj": best_sol[1],
        "QP_time": qp_time,
        "rounding_time": rounding_time,
        "sol_search_time": sol_search_time,
        "total_time": total_time,
        "eps":eps,
        "delta":delta,
        "best_sol":best_sol[0],
        "irrelevance":irrelevance,
        "lam":lam
    }
    return stat