import numpy as np
import algorithms.shared as shared
from time import time
from functools import partial
from algorithms.util import find_obj,wrapper_parallel
import multiprocessing as mp

def sim2dist(x):
    return np.sqrt(2 - 2 * x)


sim2dist_vec = np.vectorize(sim2dist)


def node_greedy_naive(W, k):
    """Naive implementation of node greedy algorithm of Borodin et al.
    Assumes that similarity is available and can be stored.
    Therefore O(n^2) storage.
    This implementation also directly minimizes the objective to be comparable to minsumsim

    Args:
        W (matrix): similarity matrix from W = A@A.T
        k (int): cardinality constraint

    Returns:
        solution: an indicator vector of 1s and 0s 
    """
    N = W.shape[0]
    S = set()
    U = set(range(N))
    # D = sim2dist_vec(W)
    d_u = {i: 0 for i in range(N)}
    first_element = np.random.choice(range(N))
    S.add(first_element)
    for v in U - S:
        d_u[v] += W[v, first_element]

    for i in range(k - 1):
        phi = {k: d_u[k] for k in U - S}
        max_u = min(phi, key=phi.get)
        S.add(max_u)
        for v in U - S:
            d_u[v] += W[max_u, v]
    indicator = np.zeros(N)
    indicator[list(S)] = 1
    return indicator


def node_greedy_large(k, starting_sol=[],A=None):
    """    Greedy Algorithm from Borodin scaled up to large N.


    Args:
        k (int): cardinality constraint
        starting_sol (list, optional): The elements that must be part of the solution. Defaults to [].
        A (matrix, optional): The data matrix, if None then taken from shared memory. Defaults to None.

    Returns:
        solution: Indicator vector of 1s and 0s
    """
    if A is None:
        A = shared.A
    # print("Greedy B: ",mp.current_process()," ",id(A))
    # print("local A: ",id(A))
    np.random.seed()
    N = A.shape[0]
    S = set()
    U = set(range(N))
    d_u = np.zeros(N)
    if len(starting_sol):
        # add starting elements to
        S.update(starting_sol)
    # if no starting elements then the first element is random
    else:
        first_element = np.random.choice(range(N))
        S.add(first_element)
    # update costs for all elements in the solution at this point
    for u in S:
        for v in U - S:
            d_u[v] += A[v] @ A[u] # compute cosine similarity on demand

    for i in range(k - len(S)):
        phi = {k: d_u[k] for k in U - S}
        # greedily pick node with least cost
        min_u = min(phi, key=phi.get)
        S.add(min_u)
        # update costs to all other elements
        for v in U - S:
            d_u[v] += A[v] @ A[min_u]

    indicator = np.zeros(N)
    indicator[list(S)] = 1
    return indicator


def node_greedy_irrelevance(k, irrelevance,A=None):
    """Greedy Algorithm from Borodin scaled up to large N and using irrelevance scores.

    Args:
        k (int): cardinality constraints
        irrelevance (vector): list of irrelevance scores for all items
        A (matrix, optional): Data matrix, will be taken from shared memory if None. Defaults to None.

    Returns:
        solution: indicator vector of 1s and 0s
    """
    if A is None:
        A = shared.A
    # print("Greedy B irr local: ",id(A))
    N = A.shape[0]
    S = set()
    U = set(range(N))
    d_u = np.zeros(N)
    S = set()
    for i in range(k):
        # scale irrelevances by half as mentioned in paper
        phi = {k: (0.5* irrelevance[k]) + d_u[k] for k in U - S}
        min_u = min(phi, key=phi.get)
        S.add(min_u)
        for v in U - S:
            d_u[v] += A[v] @ A[min_u]

    indicator = np.zeros(N)
    indicator[list(S)] = 1
    return indicator

def get_node_greedy_stat(k, starting_sol=[], time_lim=None, initial_states=10,A=None):
    """gets run time and objective values for the node greedy algorithm
    
    Args:
        k (int): cardinality requirement of solution
        starting_sol (list, optional): indices of elements that are fixed in solution. Defaults to [].
        time_lim (float, optional): the running time to try and match. Defaults to None.
        initial_states (int, optional): number of different initial states to try. Defaults to 10.
    
    Returns:
        stat: dict of the statistics
    """
    possible_sols = []

    start = time()
    b = node_greedy_large(k, starting_sol=starting_sol,A=A)
    run_time = time() - start

    assert np.allclose(sum(b), k)
    possible_sols.append(b)

    if time_lim is None:
        # repeat one less 
        repeat = initial_states - 1
    else:
        rem = time_lim - run_time
        repeat = int(np.ceil(rem / run_time))

    start = time()
    if repeat :
        p_func = partial(node_greedy_large,k=k,starting_sol=starting_sol,A=A)
        par_greedy = wrapper_parallel(p_func)
        with mp.Pool(shared.NUM_PROCESSES) as pool:
            for b in pool.map(par_greedy, range(repeat)):
                # b = Greedy_B_large(A, k, starting_sol=starting_sol)
                possible_sols.append(b)
    
    loop_time = time() - start

    start = time()
    obj = [find_obj(sol,A=A) for sol in possible_sols]
    # _,obj = find_obj_parallel(possible_sols,A)
    best_sol = min(obj)
    worst_sol = max(obj)
    avg_sol = np.mean(obj)

    sol_search_time = time() - start

    total_time = run_time + loop_time + sol_search_time

    stat = {
        "algo": "Greedy_B",
        "k": k,
        "min_obj": best_sol,
        "max_obj": worst_sol,
        "avg_obj": avg_sol,
        "single_run_time": run_time,
        "loop_time": loop_time,
        "sol_search_time": sol_search_time,
        "total_time": total_time,
        "initial_states":repeat+1,
        "best_sol":possible_sols[obj.index(best_sol)]
    }
    return stat

def get_node_greedy_irr_stat(k, irrelevance,lam=1,A=None):
    """gets run time and objective values for the node greedy algorithm
    
    Args:
        k (int): cardinality requirement of solution
        irrelevance (list): list of irrelevance scores for all items
        lam (float, optional): tradeoff parameter between irrelevance and sum of similarities. Defaults to None.
        A (matrix, optional): Data matrix, will use shared memory if None. Defaults to None.
    Returns:
        stat: dict of the statistics
    """
    # scale irrelevances by lambda
    irrelevance = irrelevance * lam
    start = time()
    b = node_greedy_irrelevance(k, irrelevance=irrelevance,A=A)
    run_time = time() - start
    assert np.allclose(sum(b), k)

    start = time()
    obj = find_obj(b,irrelevance,A=A)
    obj_time = time()-start

    total_time = run_time + obj_time
    stat = {
        "algo": "Greedy_B_irr",
        "k": k,
        "min_obj": obj,
        "run_time": run_time,
        "obj_time": obj_time,
        "loop_time":0,
        "total_time": total_time,
        "best_sol":b,
        "irrelevance":irrelevance,
        "lam":lam
    }
    return stat
