import numpy as np
import math
from functools import partial
import multiprocessing as mp
from multiprocessing.shared_memory import SharedMemory
from multiprocessing.managers import SharedMemoryManager
import random
import algorithms.shared as shared
from algorithms.util import find_obj,wrapper_parallel, NestablePool


def tree_serial(z_orig:np.ndarray):
    """A serial implementation of Srinivasan's algorithm. It internally has an 
    implementation of simplify as it reduces checking of -1 from the other function defined.

    Args:
        z_orig (np.ndarray): The un-rounded vector of shape 1xn

    Raises:
        ValueError: if there is some issues with the probability

    Returns:
        np.ndarray: The rounded vector with values \in {0,1}
    """
    z = np.copy(z_orig)
    #! Round the negative solutions that arise due to numerical instability
    negative_sols = (z<0)
    z[negative_sols] = np.round(z[negative_sols])
    # Initially all elements in fractional set
    fractional_set = list(range(len(z)))
    # run till tree level size is 1
    while len(fractional_set)>1:
        temp =[]
        if len(fractional_set)%2 != 0:
            temp.append(fractional_set[-1])
        # print(f"{len(fractional_set)=},{i=},{j=},{prob_sum=},{r=}")
        # print(f"{fractional_set=},{i=},{j=},{temp=}")
        # print(f"{len(fractional_set)} {z.sum()=}")
        for idx in range(len(fractional_set)//2):
            # select two fractional variables indices
            i,j = fractional_set[idx*2],fractional_set[idx*2+1]
            z_i,z_j = z[i],z[j]
            prob_sum = z_i+z_j
            #* set seed if reproducibility needs to be checked
            #* otherwise remove for faster code
            # random.seed(1)
            # get random number 
            r = random.random()
            # case 0 if already rounded just ignore
            if prob_sum == 0 or prob_sum == 2:
                continue 
            # case 1 
            elif prob_sum<1:
                if r <= z_i/prob_sum:
                    z[j]=0
                    z[i]=prob_sum
                    temp.append(i)
                else:
                    z[i]=0
                    z[j]=prob_sum
                    temp.append(j)
            # case 2
            elif prob_sum == 1:
                if r <= z_i:
                    z[i]=1
                    z[j]=0
                else:
                    z[i]=0
                    z[j]=1
            # case 3
            elif prob_sum<2:
                if r <= (1-z_i)/(2-prob_sum):
                    z[j]=1
                    z[i]=prob_sum-1
                    temp.append(i) 
                else:
                    z[i]=1
                    z[j]=prob_sum-1
                    temp.append(j)
            else:
                raise ValueError("Sum of probabilites exceeds 2")
        # update the set of fractional variables
        fractional_set = temp
    
    # round to ensure all are either 0 or 1
    rounded_vector = z.round()
    return rounded_vector


def simplify(i:int,j:int,z:np.ndarray):
    """The simplify method of Srinivsan's algorithm. It rounds two variables 
    given by the indices i and j in z. 

    Args:
        i (int): index of first element
        j (int): index of second element
        z (np.ndarray): the rounding vector (usually a copy)

    Raises:
        ValueError: If the probabilities sum below zero or above 2

    Returns:
        int: return index of element not rounded if any, -1 if both are rounded
    """
    z_i,z_j = z[i],z[j]
    prob_sum = z_i+z_j
    #* set seed if reproducibility needs to be checked
    #* otherwise remove for faster code
    # random.seed(1)
    # get random number 
    r = random.random()
    # print(f"{i=}{j=}{prob_sum=}{r=}")
    # case 1 
    if prob_sum == 0 or prob_sum ==1 :
        return -1
    elif prob_sum<1:
        if r <= z_i/prob_sum:
            z[j]=0
            z[i]=prob_sum
            return i
        else:
            z[i]=0
            z[j]=prob_sum
            return j
    # case 2
    elif prob_sum == 1:
        if r <= z_i:
            z[i]=1
            z[j]=0
            return None
        else:
            z[i]=0
            z[j]=1
            return None
    # case 3
    elif prob_sum<2:
        if r <= (1-z_i)/(2-prob_sum):
            z[j]=1
            z[i]=prob_sum-1
            return i 
        else:
            z[i]=1
            z[j]=prob_sum-1
            return j
    else:
        raise ValueError("Sum of probabilites exceeds 2")

def simplify_par(indices,shm_name,shape,dtype):
    """executes the simplify algorithm in a process with a shared memory object

    Args:
        indices (list of int): indices i and j to be rounded
        shm_name (str): name of the SharedMemory buffer
        shape (np.ndarray): the shape of the array
        dtype (np.dtype): the dtype of the array

    Returns:
        int: indices or -1 from simplify
    """
    i,j = indices
    # Locate the shared memory by its name
    shm = SharedMemory(shm_name)
    # Create the np.ndarray from the buffer of the shared memory
    z = np.ndarray(shape=shape, dtype=dtype, buffer=shm.buf)
    return simplify(i,j,z)
    
def tree_parallel(z:np.ndarray):
    """Runs the Srinivasan algorithm in parallel

    Args:
        z (np.ndarray): the rounding vector of shape 1xn

    Returns:
        np.ndarray: The rounded vector
    """
    shape = z.shape
    dtype = z.dtype
    # create shared memory
    shm = SharedMemory(size=z.nbytes,create=True)
    z_copy = np.ndarray(shape=shape,dtype=dtype,buffer=shm.buf)
    # copy the solution 
    z_copy[:] = z[:] 
    #! Round the negative solutions that arise due to numerical instability
    negative_sols = (z_copy<0)
    z_copy[negative_sols] = np.round(z_copy[negative_sols])
    # only select fractional variable indices from the solution
    np.random.seed()
    fractional_list = np.where((z_copy>0)&(z_copy<1))[0]
    # np.random.shuffle(fractional_list)
    # indices = [fractional_list[-1],fractional_list[-2]]
    # simplify(indices,shm.name,shape,dtype)
    # return z
    with mp.Pool(shared.NUM_PROCESSES) as pool:
        par = partial(simplify_par,shm_name = shm.name, dtype=dtype,shape=shape)
        while(len(fractional_list)>1):
            # if odd number then add last element to next list 
            if len(fractional_list)%2 !=0:
                new_fractional_list = [fractional_list[-1]]
            else:
                new_fractional_list = []
            # calculate optimal chunksize
            chunksize = max(len(fractional_list)//(2*shared.NUM_PROCESSES),1)
            # chunksize=100
            # print(len(fractional_list),chunksize,z_copy.sum())
            # zip every two consecutive elements in the indices list
            successive_groups = zip(fractional_list[::2],fractional_list[1::2])
            #* use imap if want to compare to sequential version
            # Otherwise can switch to unordered for faster running 
            for not_rounded in pool.imap(par,successive_groups,chunksize=chunksize):
                if not_rounded is not None:
                    new_fractional_list.append(not_rounded)
            fractional_list = new_fractional_list
    # round to ensure all are either 0 or 1
    rounded_vector = z_copy.round()
    shm.close()
    shm.unlink()
    return rounded_vector
            

def randomRounding(z):
    """Functions that rounds the probability vector.

    Args:
        z (vector): probability vector from QP solver

    Returns:
        x: rounded vector with only 0s or 1s
    """
    np.random.seed()
    r = np.random.rand(*z.shape)
    x = (r <= z).astype(int)
    return x

def get_single_feasible(z, k):
    """performs randomized rounding till a feasible solution is obtained.
    A feasible solution has exactly k items selected.

    Args:
        z (vector): solution from QP solver
        k (int): cardinality constraint

    Returns:
        vector: an indicator vector that is a feasible solutions
    """
    feasible = False
    while not feasible:
        solution = randomRounding(z)
        if sum(solution) == k:
            feasible = True
    return solution

def randomRounding_alt(z,shape):
    np.random.seed()
    r = np.random.random_sample(shape)
    x = (r <= z).astype(int)
    return x

def get_single_feasible_alt(z,k,delta=1e-2):
    num = math.ceil(math.sqrt(2*math.pi*k)*math.log(1/delta))
    shape = (num,z.shape[0])
    while True:
        x = randomRounding_alt(z,shape)
        idxs = np.nonzero(x.sum(axis=1)==k)[0]
        # if no feasible solution found try again
        if not len(idxs):
            continue
        else:
            return x[idxs[0]]
    




def get_eps_delta_solutions(z, k, eps=1e-2, delta=1e-2, parallel=False,rounding_function=get_single_feasible_alt):
    """Function that performs randomized rounding multiple times in parallel.
    The number of times is based on the quality of solution required.
    This is determined by parameters eps and delta.
    Num of solution = log(1/delta)/log(1+eps) as described in the paper

    Args:
        z (vector): solution vector from the QP
        k (int): cardinality constraint
        eps (float, optional): How far the solution should be from the optimal. Defaults to 1e-2.
        delta (float, optional): What probability the set of feasible solutions should have a viable solution. Defaults to 1e-2.
        parallel (bool, optional): Whether to find feasible solutions in parallel. Defaults to False.

    Returns:
        list of indicator vectors which are feasible solutions
    """
    num = math.ceil(math.log(1 / delta) / eps)
    # print(f"Will find {num} feasible solutions")
    l = []
    if not parallel:

        for _ in range(int(num)):
            l.append(rounding_function(z, k))
    else:
        # set z and k
        par = partial(rounding_function, z=z, k=k)
        # wrap as now we have no args
        wrap = wrapper_parallel(par)
        with mp.Pool(processes=shared.NUM_PROCESSES) as pool:
            for sol in pool.imap_unordered(wrap, range(int(num))):
                l.append(sol)

    return l

def get_tree_eps_delta_solutions(z, k, eps=1e-2, delta=1e-2, parallel=False,rounding_function=tree_serial):
    """Function that performs randomized rounding multiple times in parallel.
    The number of times is based on the quality of solution required.
    This is determined by parameters eps and delta.
    Num of solution = log(1/delta)/log(1+eps) as described in the paper

    Args:
        z (vector): solution vector from the QP
        k (int): cardinality constraint
        eps (float, optional): How far the solution should be from the optimal. Defaults to 1e-2.
        delta (float, optional): What probability the set of feasible solutions should have a viable solution. Defaults to 1e-2.
        parallel (bool, optional): Whether to find feasible solutions in parallel. Defaults to False.

    Returns:
        list of indicator vectors which are feasible solutions
    """
    num = np.ceil(np.log(1 / delta) / eps)
    # print(f"Will find {num} feasible solutions")
    l = []
    if not parallel:

        for _ in range(int(num)):
            l.append(rounding_function(z))
    else:
        # set z
        par = partial(rounding_function, z_orig=z)
        # wrap as now we have no args
        wrap = wrapper_parallel(par)
        with mp.Pool(processes=shared.NUM_PROCESSES) as pool:
            for sol in pool.imap_unordered(wrap, range(int(num))):
                l.append(sol)

    return l

