from pathlib import Path
import algorithms.shared as shared
import numpy as np
from algorithms.util import preprocess_data
from tqdm.auto import tqdm
import multiprocessing as mp
from experiments.utils import import_obj
import json


project_root = Path(__file__).resolve().parent.parent

def dict_2_string(d):
    """Converts a dictionary of stats into a string for
    the csv file
    
    Args:
        d (dict): dictionary of the statistics
    
    Raises:
        UserWarning: if there is unknown algorithm being used
    
    Returns:
        stat_string: row of csv with columns: 
                     algorithm, k, min, max, avg, t1, t2, t3, total_time
    """ 
    strings = []
    algo = d["algo"]
    if algo == "MinSimSum":
        # min=max=avg
        strings.append(d["algo"])
        strings.append(d["k"])
        strings.extend([d["min_obj"]] * 3)
        strings.append(d["QP_time"])
        strings.append(d["rounding_time"])
        strings.append(d["sol_search_time"])

    elif algo == "Greedy_B" or algo == "Random_baseline":
        algo = f"{algo}_init_{d['initial_states']}"
        strings.append(algo)
        strings.append(d["k"])
        strings.append(d["min_obj"])
        strings.append(d["max_obj"])
        strings.append(d["avg_obj"])
        strings.append(d["single_run_time"])
        strings.append(d["loop_time"])
        strings.append(d["sol_search_time"])
    
    elif algo =="Greedy_A":
        strings.append(d["algo"])
        strings.append(d["k"])
        strings.extend([d["min_obj"]] * 3)
        strings.append(d["w_creation_time"])
        strings.append(d["run_time"])
        strings.append(d["obj_time"])

    else:
        print("Unknown algorithm please check code")
        raise UserWarning

    strings.append(d["total_time"])

    stat_string = ",".join([str(i) for i in strings])
    stat_string += "\n"
    return stat_string



def execute_lambda_0(args,dataset_dict,algorithm_dict,overwrite_mode="None"):

    shared.update_num_proc(args.num_processes)
    # location of the results folder
    results_folder = project_root /"data"/ "results" / "lambda_0" / dataset_dict['dataset']
    # create if it doesn't exist
    if not results_folder.exists():
        results_folder.mkdir(parents=True,exist_ok=True)

    kwargs= algorithm_dict['kwargs']
    stat_fn = import_obj(algorithm_dict['stat_fn'])
    
    if "sparse" in kwargs:
        kwargs["sparse"]=dataset_dict["sparse"]
    
    ks = [2**i for i in range(2,8)]

    with np.load(dataset_dict["npz_location"]) as data:

        results_file = results_folder / f"{algorithm_dict['name']}.csv"
        file_exists = results_file.exists()
        
        if overwrite_mode == "None" and file_exists:
            raise ValueError(f"Result file for dataset {dataset_dict['dataset']} and algorithm {algorithm_dict['name']} already exists. Change overwrite mode to overwrite or append to gather new results")
        elif overwrite_mode == "overwrite":
            mode ="w"
            file_exists = False
        elif overwrite_mode == "append":
            mode="a"
        # open the results file
        with open( results_file, mode=mode, encoding="utf-8") as fp:
            if not file_exists:
                fp.write("algorithm,k,min_obj,max_obj,avg_obj,t_1,t_2,t_3,t_total\n")
            # get the data matrix
            shared.update_W(None)
            A = data[dataset_dict['dataset']]
            shared.update_A(A)
            rel, qid, A = preprocess_data(dataset_dict["letor"])
            shared.update_A(A)
            print(f"Starting {dataset_dict['dataset']} dataset: {A.shape}\n")
            # how many times to repeat the algorithm for each k
            repeats_tqdm = tqdm(range(algorithm_dict['repeats']))
            for i in repeats_tqdm:
                repeats_tqdm.set_description(f"iteration {i+1}")    
                k_tqdm = tqdm(ks)
                for k in k_tqdm:
                    k_tqdm.set_description(f"k = {k}")
                    kwargs['k']=k
                    stats = stat_fn(**kwargs)
                    fp.write(dict_2_string(stats))
                    fp.flush()  