from algorithms.util import preprocess_data
import algorithms.shared as shared
import numpy as np
from collections import Counter
from tqdm.auto import tqdm
from experiments.utils import import_obj
from pathlib import Path

project_root = Path(__file__).resolve().parent.parent


def dict_2_string(d):
    """Converts a dictionary of stats into a string for
    the csv file
    
    Args:
        d (dict): dictionary of the statistics
    
    Raises:
        UserWarning: if there is unknown algorithm being used
    
    Returns:
        stat_string: row of csv with columns: 
                     algorithm, k, query, lam, min, max, avg, t1, t2, t3, total_time
    """ 
    strings = []
    algo = d["algo"]
    strings.append(algo)
    strings.append(d["k"])
    strings.append(d["query"])
    strings.append(d["lam"])

    if algo == "MinSimSum":
        strings.append(d["min_obj"])
        strings.append(d["QP_time"])
        strings.append(d["rounding_time"])
        strings.append(d["sol_search_time"])

    elif algo == "Greedy_B_irr" or algo == 'Greedy_A_irr':
        strings.append(d["min_obj"])
        strings.append(d["run_time"])
        strings.append(d["loop_time"])
        strings.append(d["obj_time"])
    else:
        print("Unknown algorithm please check code")
        raise UserWarning

    strings.append(d["total_time"])
    stat_string = ",".join([str(i) for i in strings])
    stat_string += "\n"
    return stat_string

def process_query(query,k,lam,stat_fn):
    
    # get the subset of the data related to query
    A_query = shared.A[shared.qid==query]
    # get respective relevance scores 
    relevances = shared.rel[shared.qid==query]
    # normalize relevance from {0,1,2,3,4} to {0.2,0.4,0.6,0.8,1} for Istella
    # or {0,1,2} to {0.33,0.66,1} for LETOR
    relevances = (relevances+1)/len(np.unique(shared.rel))
    # convert them into irrelevance
    irrelevances = 1+np.log((1/relevances))

    # gather stats
    stats = stat_fn(k=k,irrelevance=irrelevances,A=A_query,lam=lam)
    # add query to stats
    stats['query'] = query
    # create string
    stat_str = dict_2_string(stats)

    return stat_str

def execute_lambda_query(args,data_dict,alg_dict,overwrite_mode="None"):
    # location of the results folder
    results_folder = project_root /"data"/ "results" / "lambda_query" / data_dict['dataset']
    # create if it doesn't exist
    if not results_folder.exists():
        results_folder.mkdir(parents=True,exist_ok=True)

    # the range for k
    ks = [2**i for i in range(2,7)]

    results_file = results_folder / f"{alg_dict['name']}.csv"
    file_exists = results_file.exists()
    
    if overwrite_mode == "None" and file_exists:
        raise ValueError(f"Result file for dataset {data_dict['dataset']} and algorithm {alg_dict['name']} already exists. Change overwrite mode to overwrite or append to gather new results")
    elif overwrite_mode == "overwrite":
        mode ="w"
        file_exists = False
    elif overwrite_mode == "append":
        mode="a"
    else:
        mode="a"
    
    # load the algorithm 
    stat_fn = import_obj(alg_dict['stat_fn'])

    print(f"Loading {data_dict['dataset']} dataset")
    X = np.load(data_dict['npz_location'])[data_dict['dataset']]
    shared.update_A(X)
    rel,qid,X = preprocess_data(letor=data_dict['letor'])
    shared.update_A(X)
    shared.rel = rel
    shared.qid = qid
    print("Done")

    counts = Counter(qid)
    # gather only the queries with the most documents 
    biggest_query_size = max(counts.values())
    # keep a list of the query ids with largest number of docs
    large_qid = [ q for q in np.unique(qid) if counts[q]==biggest_query_size ]

    print(f"Found {len(large_qid)} queries with {biggest_query_size} documents each")

    with open(results_file,mode=mode,encoding='utf-8') as fp:
        if not file_exists:
            fp.write("algorithm,k,query,lam,min_obj,t_1,t_2,t_3,t_total\n")
        query_tqdm = tqdm(large_qid[:100])
        for query in query_tqdm:
            query_tqdm.set_description(f"Query no={query}")
            shared.update_W(None)
            k_tqdm = tqdm(ks)
            for k in k_tqdm:
                k_tqdm.set_description(f"k={k}")
                lam_tqdm = tqdm(args.lambdas)
                for lam in lam_tqdm:
                    lam_tqdm.set_description(f"lambda={lam:.1f}")    
                    stat_str = process_query(query,k,lam,stat_fn)
                    fp.write(stat_str)
                    fp.flush()



