import numpy as np
import multiprocessing as mp
from functools import partial
import algorithms.shared as shared
from pathlib import Path
from tqdm import tqdm
from time import time
from algorithms.util import preprocess_data,MyPool,find_obj
from experiments.utils import import_obj

project_root = Path(__file__).resolve().parent.parent

def process_sample(sample,k,algorithm,kwargs):
    """A function that gets the solution for a sample.
    A sample is a random subset of the items.

    Args:
        sample (array): A list of rows that corresponds to the items in the sample
        k (int): cardinality constraint
        algorithm (function): The get_stat function of the algorithm in consideration
        kwargs (dict): the key word arguments for the algorithm

    Returns:
        array: the ids of the sample that are part of the solution
    """
    # get the subset of the rows as per the sample
    A_sample = shared.A[sample]
    # get the solution from the algorithm
    sol = algorithm(k=k,A=A_sample,**kwargs)["best_sol"]
    assert sol.shape== sample.shape
    # get the actual ids of the sample part of the solution
    ids = sample[sol==1]
    return ids

def subsample(k,N,alg,kwargs):
    start = time()
    solutions = set()
    # create a partial function with the values of k and algorithm set
    par = partial(process_sample,k=k,algorithm=alg,kwargs=kwargs)
    # reduce the number of cores as each algorithm finds a solution using more threads
    cores = int(np.floor(np.sqrt(mp.cpu_count())))
    shared.update_num_proc(cores)
    # start top level process with sqrt(cpus+1)
    # this way the total number of processes will be sqrt(cpus+1)x sqrt(cpus)
    with MyPool(processes=cores+1) as pool:    
        for s in pool.map(par,shared.samples):
            solutions.update(s)
    
    # gather solution from all samples 
    solutions = np.array(list(solutions))
    # take a subset of those solutions 
    A_subsample = shared.A[solutions]
    final_sol = alg(k=k,A=A_subsample,**kwargs)["best_sol"]
    assert final_sol.shape == solutions.shape
    final_x = np.zeros(N)
    # get final solution as items in the last soltition
    ids = solutions[final_sol==1]
    final_x[ids]=1
    # find the objective of the final solution
    best_obj = (find_obj(final_x))
    run_time = time()-start
    return best_obj,run_time

def execute_subsampling(args,data_dict,alg_dict,overwrite_mode="None"):

    # location of the results folder
    results_folder = project_root /"data"/ "results" / "subsampling" / data_dict['dataset']
    # create if it doesn't exist
    if not results_folder.exists():
        results_folder.mkdir(parents=True,exist_ok=True)

    results_file = results_folder / f"{alg_dict['name']}.csv"
    file_exists = results_file.exists()
    
    if overwrite_mode == "None" and file_exists:
        raise ValueError(f"Result file for dataset {data_dict['dataset']} and algorithm {alg_dict['name']} already exists. Change overwrite mode to overwrite or append to gather new results")
    elif overwrite_mode == "overwrite":
        mode ="w"
        file_exists = False
    elif overwrite_mode == "append":
        mode="a"
    else:
        mode="a"    

    # load the algorithm 
    stat_fn = import_obj(alg_dict['stat_fn'])
    kwargs = alg_dict["kwargs"]

    print(f"Loading {data_dict['dataset']} dataset....")
    A = np.load(data_dict["npz_location"])[data_dict["dataset"]]
    shared.update_A(A)
    rel,qid,A = preprocess_data(letor=data_dict["letor"])
    shared.update_A(A)
    N = A.shape[0]
    print("Done")

    n_samples_range = args.n_samples
    ks = [2**i for i in range(2,7)]
    # size of each sample
    size = int(np.sqrt(N))

    with open(results_file, mode=mode,encoding='utf-8') as fp:
        if not file_exists:
            fp.write("algorithm,n_samples,k,min_obj,run_time\n")
        rep_tqdm = tqdm(range(args.repetitions)) 
        for r in rep_tqdm :
            rep_tqdm.set_description(f"Iteration={r+1}")
            samples_tqdm = tqdm(n_samples_range)             
            for n_samples in samples_tqdm:
                samples_tqdm.set_description(f"Number of samples={n_samples}")
                # generate samples 
                samples = np.array([np.random.choice(N,size,replace=False) for _ in range(n_samples)])
                shared.update_samples(samples)
                k_tqdm = tqdm(ks)
                for k in k_tqdm:
                    k_tqdm.set_description(f"k={k}")
                    best_obj,run_time = subsample(k=k,N=N,alg=stat_fn,kwargs=kwargs)
                    fp.write(f"{alg_dict['name']},{n_samples},{k},{best_obj},{run_time}\n")
                    fp.flush()




