import numpy as np
from pathlib import Path
from tqdm import tqdm

"""The Istella-X dataset is large and not processed like the Microsoft LETOR datasets.
We convert the raw istellla-x files into a 'MIN' version as described in 
https://www.microsoft.com/en-us/research/project/letor-learning-rank-information-retrieval/#!letor-4-0

"""

def convert_tok(s):
    """converts token in each columns of letor data.
    letor data has the format:
    1 qid:1234 1:00 2:22 ... #comment
    This function splits the qid:1234 style token and returns the value
    Args:
        s (string): input token string

    Returns:
        float: the value in the token
    """
    return float(s.split(":")[-1])

def inf_to_min(A):
    # get relevance scores
    rel = A[:,0].astype(np.int8)
    # get query ids
    qid = A[:,1].astype(int)
    # matrix of feature vectors
    A = A[:,2:] 

    # null mask (stored as inf)
    null_mask = A==np.max(A)
    uniq_qid = np.unique(qid)
    
    valid_quid =[]
    print("Starting qid loop")
    for q in tqdm(uniq_qid):
        # mask for the rows belonging to the query
        qid_mask = (qid==q)
        
        # get indices where the null values are for a qid
        row_ind,col_ind = np.where(null_mask[qid_mask])
        
        #! There are times when the min feature in a query is also inf
        # find the smallest value for each feature in the query
        min_vec = np.min(A[qid_mask],axis=0)
        # if there are still inf values then just skip the quid
        if np.any(min_vec > 1e308):
            continue    

        # set the null features with min for that query
        A[null_mask & qid_mask[:,np.newaxis] ] = min_vec[col_ind]
        valid_quid.append(q)

    valid_mask = np.isin(qid,valid_quid)
    A = np.column_stack((rel,qid,A))    
    return A[valid_mask]

def istella_to_npz(name,raw_data_dir,save_dir):

    istella_letor_file = raw_data_dir/"istella-X"/f"{name}.txt"
    features = 220
    print(f"Processing Istella {name} to npz")
    # cannot use pyltr as the datafiles are too large
    # instead load the files using numpy and converters

    A = np.loadtxt(
        istella_letor_file,
        encoding='latin-1', 
        comments="#", 
        converters={ # convert token for all features like qid:1234
            # dimension = relevance score + qid + features columns
            i:convert_tok for i in range(1,features+2) 
            }
        )    
    
    kwargs = {name:A}
    
    print("Storing File. Might take a while...")
    save_file = save_dir/ f"istella_{name}.npz"
    np.savez_compressed(save_file,**kwargs)

def istella_to_min_npz(name,raw_data_dir,save_dir):
    if not (raw_data_dir/"istella-X").exists():
        raise FileNotFoundError(f"Istella-X raw data not found at {raw_data_dir/'istella-X'}, please download then try again")
    
    if name != "test" and name != "vali":
        raise ValueError(f"Istella name {name} is not valid, please enter either test or vali")

    istella_npz_file = save_dir/ f"istella_{name}.npz"
    if not (istella_npz_file).exists():
        istella_to_npz(name,raw_data_dir,save_dir)
    else :
        print(f"Istella {name} npz exists, skipping to next stage")

    min_file = save_dir/f"istella_{name}_min.npz"
    print("Replacing NULLs with MIN in npz file")
    with np.load(istella_npz_file) as data:
        A = data[data.files[0]]
        A = inf_to_min(A)
        kwargs = {istella_npz_file.stem: A}
        print("Saving MIN file. Might take a while...")
        np.savez_compressed(min_file,**kwargs)


